#!flask/bin/python
import logging
import os

import flask
from flask import Flask
import simplejson as json

from flask import request
from werkzeug.exceptions import NotFound
from apispec import APISpec
from apispec_webframeworks.flask import FlaskPlugin
from apispec.ext.marshmallow import MarshmallowPlugin
from opentelemetry import trace
from opentelemetry.instrumentation.logging import LoggingInstrumentor
from opentelemetry.instrumentation.flask import FlaskInstrumentor
from opentelemetry.instrumentation.requests import RequestsInstrumentor
from opentelemetry.sdk.trace import TracerProvider

from services.auth_ctx_service import AuthContextService
from services.helper_service import HelperService
from services.login_service import LoginService
from services.cache_service import CacheService
from services.xcsrf_service import XCSRFService
from services.access_service import AccessService
from models.user_context import UserContext
from werkzeug.exceptions import Forbidden
from services.xcsrf_service import CSRF_TOKEN_HEADER
from models.api.login_request import LoginRequest
from models.api.update_rbac_request import UpdateRbacRequest
from models.api.switch_orgs_request import SwitchOrgsRequest

from util import org_util

# APISpec for swagger documentation
spec = APISpec(
    title='psc.auth',
    version='1.0.0',
    openapi_version='2.0',
    plugins=[
        FlaskPlugin(),
        MarshmallowPlugin(),
    ],
)

app = Flask(__name__)
# Set the tracer provider and instrument Flask, Logging and Requests
trace.set_tracer_provider(TracerProvider())
FlaskInstrumentor().instrument_app(app)
RequestsInstrumentor().instrument()
log_fmt = '%(asctime)s [%(process)d] [%(levelname)s] [%(name)s] \
[trace_id=%(otelTraceID)s span_id=%(otelSpanID)s] - %(message)s'
date_fmt = '[%Y-%m-%d %H:%M:%S %z]'
LoggingInstrumentor().instrument(logging_format=log_fmt, date_format=date_fmt)


if __name__ != '__main__':
    # Setup logging, by attaching Flask's logs to the Gunicorn's handler
    # and reusing its log level
    # Gunicorn handlers are also set to root logger's handlers
    gunicorn_logger = logging.getLogger('gunicorn.error')
    gunicorn_logger.handlers = logging.getLogger().handlers
    app.logger.handlers = gunicorn_logger.handlers
    app.logger.setLevel(gunicorn_logger.level)
    app.logger.propagate = False
else:
    app.logger.setLevel(os.getenv("LOG_LEVEL", "INFO"))

xsrf = XCSRFService(app)
auth_ctx_service = AuthContextService(app, xsrf)
helper_service = HelperService()
login_service = LoginService(app, xsrf)
cache_service = CacheService(app)
access_service = AccessService(app, helper_service)

try:
    from http.cookies import SimpleCookie
except ImportError:
    from Cookie import SimpleCookie


@app.before_request
def before_request():
    headers = str(request.headers).replace('\r', '').rstrip().split('\n')
    body = request.get_data().decode('utf-8')
    app.logger.debug('{} {}; Headers: {}; Body: {}'.format(request.method, request.path, headers, body))


@app.after_request
def after_request(response):
    app.logger.debug('End of request {} {}; Status: {}'.format(request.method,
                                                               request.path,
                                                               response.status_code))
    return response


@xsrf.exempt
@app.route('/auth/v1/sessions/current', methods=['GET', 'DELETE', 'PUT'])
def authenticated():
    """Add, remove, or modify sessions in the cache
            ---
            get:
                description: Check if a request is authenticated by checking for an existing session in the cache
                responses:
                    200:
                        description: The request is authenticated
                    401:
                        description: The request is not authenticated
            delete:
                description: Remove a session from the cache
                responses:
                    200:
                        description: The session was successfully removed if it existed
            put:
                description: >
                    Update a session's target org, using the assume_identity backend method or the masquerade
                    backend method.  By default, assume_identity method is used unless the masquerade argument in the
                    url query string is set to 'true': /auth/v1/sessions/current?masquerade=true
                parameters:
                    - in: body
                      schema: SwitchOrgsRequest
                    - in: query
                      type: string
                      name: backend-method
                responses:
                    200:
                        description: The target org was successfully changed
    """
    if request.method == "GET":
        return check_authenticated()
    elif request.method == "DELETE":
        return logout()
    elif request.method == "PUT":
        try:
            org_id = org_util.get_org_id_from_request(request)
        except ValueError as err:
            response = app.make_response((
                json.dumps({
                    "message": err.args[0],
                    "success": False}),
                400
            ))
            response.headers['X-Processed-By'] = "psc.auth"
            return response

        if request.args.get('masquerade') == "true":
            appservices_url_format = '{gateway_url}/appservices/v1/delegations/masquerade/{org_id}'
        else:
            appservices_url_format = '{gateway_url}/appservices/v5/orgs/{org_id}/assume-identity'
        return login_service.switch_org(request, int(org_id), appservices_url_format)


@xsrf.exempt
@app.route('/auth/v1/sessions', methods=['POST'])
def login_v1():
    """Add, remove, or modify sessions in the cache
        ---
        post:
            parameters:
                - in: body
                  schema: LoginRequest
            description: Create a new session in the cache
            responses:
                200:
                    description: There were no errors and a session was created if the credentials were validated
    """
    return login()


@xsrf.exempt
@app.route('/auth/v1/sessions/saml', methods=['POST'])
def login_saml_v1():
    """Add a session to the cache using SAML
       ---
       post:
           description: Create a new session in the cache
           responses:
               200:
                   description: There were no errors and a session was created if the credentials were validated
    """
    return login_saml()


@app.route('/auth/v1/orgs/<string:org_key>/users/<int:auth_id>/reset', methods=['PUT'])
def reset_user_sessions(org_key=None, auth_id=None):
    """Modify all sessions that belong to a principal except the current one
        ---
        put:
            parameters:
                - in: path
                  name: org_key
                  required: true
                  type: string
                - in: path
                  name: auth_id
                  required: true
                  type: integer
            description: Modify user sessions so that only current session in call context remains active
            responses:
                200:
                    description: >
                        The cache values are successfully updated if there were other sessions besides the current one.
                        If there are no additional sessions besides the current one, the number of sessions invalidated
                        is 0.
    """
    return reset_sessions(org_key, auth_id)


@app.route('/auth/v1/orgs/<string:org_key>/users/<int:auth_id>/sessions', methods=['DELETE', 'PUT'])
@app.route('/auth/v1/orgs/<string:org_key>/connectors/<string:auth_id>/sessions', methods=['DELETE', 'PUT'])
def update_user_sessions(org_key=None, auth_id=None):
    """Modify all sessions that belong to a principal
        ---
        put:
            parameters:
                - in: body
                  schema: UpdateRbacRequest
                - in: path
                  name: org_key
                  required: true
                  type: string
                - in: path
                  name: auth_id
                  required: true
                  type: integer
            description: Update the RBAC key of a user or connector's sessions
            responses:
                200:
                    description: >
                        The cache value was successfully updated if a stale value existed in the cache.  The caller must
                        have permission to update principals with the old RBAC key as well as the new RBAC key.  If
                        there was no cached RBAC key, the cache was not updated and the value of the response's
                        "success" field is "false"
                403:
                    description: >
                        The caller is forbidden from updating principals with either the old RBAC key or the new
                        RBAC key, or both
        delete:
            parameters:
                - in: path
                  name: org_key
                  required: true
                  type: string
                - in: path
                  name: auth_id
                  required: true
                  type: integer
            description: Delete all sessions belonging to a principal
            responses:
                200:
                    description: Success
                403:
                    description: The caller is forbidden
    """
    if request.method == 'PUT':
        return update_session_rbac_key(org_key, auth_id)
    if request.method == 'DELETE':
        return delete_sessions(org_key, auth_id)


@xsrf.exempt
@app.route('/auth/v1/swagger')
def get_swagger_spec():
    return app.make_response((json.dumps(spec.to_dict()), 200))


@xsrf.exempt
@app.route('/isauthenticated')
def is_authenticated():
    return check_authenticated()


@app.route('/')
def check_authenticated():
    try:
        req_ctx = auth_ctx_service.get_request_context(request)
        session_ctx = auth_ctx_service.get_session_context(request, request_ctx=req_ctx)
        response = auth_ctx_service.make_response(request, req_ctx, session_ctx)
        if HelperService.is_csrf_exempt_authenticated_request(request) and session_ctx.authenticated:
            response.headers[CSRF_TOKEN_HEADER] = session_ctx.xcsrf
    except Forbidden as e:
        response = app.make_response(
            (
                json.dumps({
                    "message": e.description,
                    "error_code": 'FORBIDDEN'
                }),
                403
            )
        )
    return response


@xsrf.exempt
@app.route('/login', methods=['POST'])
def login():
    return login_service.login(request, '{gateway_url}/appservices/v5/admin/login', request.data)


@xsrf.exempt
@app.route('/login/saml/consume', methods=['POST'])
@app.route('/samlLogin/consume/<saml_code>', methods=['POST'])
@app.route('/samlLogin/consume', methods=['POST'])
def login_saml(saml_code=None):
    incoming = {
        key: value[0] if len(value) == 1 else value
        for key, value in request.form.items()
    }
    return login_service.login(
        request,
        '{gateway_url}/appservices/v5/admin/login/saml/consume',
        json.dumps(incoming),
        saml=True
    )


@app.route('/userInfo', methods=['GET'])
def user_info():
    req_ctx = auth_ctx_service.get_request_context(request)
    session_ctx = auth_ctx_service.get_session_context(request, request_ctx=req_ctx)
    return json.dumps(session_ctx.user_info), 200, {'X-Processed-By': "psc.auth"}


@xsrf.exempt
@app.route('/logout', methods=['GET'])
def logout():
    call_ctx = auth_ctx_service.get_call_context(request)
    cache_service.purge_user_cache(call_ctx.session_id, call_ctx.user_id or call_ctx.connector_id)
    return json.dumps({"success": True}), 200, {'X-Processed-By': "psc.auth"}


def delete_sessions(org_key=None, auth_id=None):
    req_ctx = auth_ctx_service.get_request_context(request)
    call_ctx = auth_ctx_service.get_call_context(request)

    if not call_ctx.is_api_call() and not call_ctx.is_session_call():
        raise NotFound("session not found")

    session_id = None
    if call_ctx.auth_token:
        session_id = auth_ctx_service.get_session_id_from_auth_token(call_ctx.auth_token)
    else:
        session_id = call_ctx.session_id

    session_ctx = cache_service.get_session_context(session_id)

    if session_ctx.origin_env != req_ctx.origin_env:
        resp_data = {
            "message": "request context origin environment mismatch",
            "error_code": 'UNAUTHORIZED'
        }
        resp_code = 401
    else:
        try:
            permissions = [{
                "name": "csr.auth.session",
                "actions": ["DELETE"]
            }]
            access_service.get_demand(request, req_ctx.gateway_ip, session_ctx.user_context, org_key, None, permissions)
            num_invalidated = cache_service.evict_sessions(auth_id)

            resp_data = {
                "success": True,
                "invalidated_sessions": num_invalidated
            }
            resp_code = 200
        except Forbidden as e:
            resp_data = {
                "message": e.description,
                "error_code": 'FORBIDDEN'
            }
            resp_code = 403
    return app.make_response((json.dumps(resp_data), resp_code))


def reset_sessions(org_key=None, auth_id=None):
    call_ctx = auth_ctx_service.get_call_context(request)

    # This is to protect calls from internal endpoint
    if not call_ctx.session_id:
        raise NotFound("session not found")

    num_invalidated = cache_service.reset_sessions(auth_id, org_key, call_ctx.session_id)

    resp_data = {
        "success": True,
        "invalidated_sessions": num_invalidated
    }

    return app.make_response((json.dumps(resp_data)))


def update_session_rbac_key(org_key=None, auth_id=None):
    req_ctx = auth_ctx_service.get_request_context(request)
    user_ctx = UserContext.parse({"user_context": json.loads(request.headers['X-User-Context'])})

    if not user_ctx:
        raise NotFound("session not found")

    req_body = json.loads(request.data)
    new_claim = req_body['rbac_key']
    old_claim = cache_service.get_session_rbac_key(auth_id)

    # Only update if there is a stale value in the cache
    try:
        permissions = [
            {
                "name": "org.users",
                "actions": ["UPDATE"]
            }
        ]
        claims = [new_claim, old_claim] if old_claim else [new_claim]
        access_service.get_demand(request, req_ctx.gateway_ip, user_ctx, org_key, claims, permissions)
        if old_claim:
            # The rbac key is cached, and must be updated
            cache_service.update_session_rbac_key(auth_id, new_claim)
            response = json.dumps({"success": True}), 200, {'X-Processed-By': "psc.auth"}
        else:
            # The cached rbac key does not need to be updated as there is no stale value
            response = json.dumps({"success": False}), 200, {'X-Processed-By': "psc.auth"}

    except Forbidden as e:
        response = app.make_response(
            (
                json.dumps({
                    "message": e.description,
                    "error_code": 'FORBIDDEN'
                }),
                403
            )
        )
    return response


@app.route('/appservices/v1/delegations/masquerade/<org_id>', methods=['POST'])
def post_delegation(org_id):
    return login_service.switch_org(request, int(org_id),
                                    '{gateway_url}/appservices/v1/delegations/masquerade/{org_id}')


@app.route('/appservices/v5/orgs/<org_id>/assume-identity', methods=['POST'])
def assume_identity(org_id):
    return login_service.switch_org(request, int(org_id), '{gateway_url}/appservices/v5/orgs/{org_id}/assume-identity')


@app.route('/auth/v1/users/<email_id>/context', methods=['GET'])
def get_user_context_headers(email_id):
    app.logger.info("Get Context API: Call to /auth/v1/users/<email_id>/context endpoint for user {}".format(email_id))
    return login_service.verify_user_email(request, email_id, '{gateway_url}/adminServices/v5/users/{email}')


@xsrf.exempt
@app.route('/auth/v1/health_check')
@app.route('/healthCheck')
def health_check():
    return json.dumps({
        "message": "auth service is running",
        "hash": os.getenv("HASH", "")
    })


@app.errorhandler(404)
def not_found(e):
    return flask.jsonify(error_code="NOT_FOUND", message=str(e.description)), 404


if __name__ == '__main__':
    app.run(debug=True, host='0.0.0.0')

with app.test_request_context():
    spec.components.schema('login_request', schema=LoginRequest)
    spec.components.schema('update_rbac_request', schema=UpdateRbacRequest)
    spec.components.schema('switch_orgs_request', schema=SwitchOrgsRequest)
    spec.path(view=authenticated)
    spec.path(view=login_saml_v1)
    spec.path(view=authenticated)
    spec.path(view=login_v1)
    spec.path(view=update_user_sessions)
    spec.path(view=reset_user_sessions)
