import time

import requests
import simplejson as json

from services.helper_service import HelperService
from werkzeug.exceptions import NotFound, Unauthorized, Forbidden


class AccessService(object):

    def __init__(self, app=None, helper_service=None):
        self.app = app
        self.helper_service = helper_service or HelperService()

    def get_session_rbac_key(self, gateway_ip, headers, org_key, user_id, connector_id, connector_type):
        headers = dict(headers)
        headers.pop('Content-Length', None)

        if not user_id and not connector_id:
            return None

        if user_id:
            url = '{gateway_url}/access/v2/orgs/{org_key}/users/{user_id}/claims'
            url = url.format(
                gateway_url=self.helper_service.get_gateway_url(gateway_ip),
                org_key=org_key,
                user_id=user_id,
                connector_id=connector_id,
                connector_type=connector_type
            )
        else:
            url = '{gateway_url}/access/v2/orgs/{org_key}/connectors/{connector_id}/claims?type={connector_type}'
            url = url.format(
                gateway_url=self.helper_service.get_gateway_url(gateway_ip),
                org_key=org_key,
                user_id=user_id,
                connector_id=connector_id,
                connector_type=connector_type
            )

        method = "GET"
        self.app.logger.debug(f"Request {method} {url}")
        req_start = time.time()
        rbac_response = requests.request(
            method=method,
            url=url,
            headers=headers
        )
        req_rtt = time.time() - req_start
        self.app.logger.debug(f"Response status code {rbac_response.status_code}; Took {req_rtt:.4f}s")

        rbac_key = None
        if rbac_response.status_code == 200:
            try:
                rbac_key = json.loads(rbac_response.text)["claims"]
            except:
                pass

        if rbac_key:
            return rbac_key

        self.app.logger.warning(
            "Failed to retrieve rbac key from access service. "
            "org_key: {org_key}, "
            "user_id: {user_id}, "
            "connector_id: {connector_id}, connector_type: {connector_type},"
            "response_code: {rbac_response_code}, response_text: {rbac_response_text}".format(
                org_key=org_key,
                user_id=user_id or "",
                connector_id=connector_id or "",
                connector_type=connector_type or "",
                rbac_response_code=rbac_response.status_code,
                rbac_response_text=rbac_response.text
            )
        )
        raise Forbidden("Unable to retrieve rbac key")

    def get_demand(self, request, gateway_ip, user_context, org_key, rbac_keys, permissions, env=None):
        entities = {
            "org": org_key,
            "roles": rbac_keys
        }
        caller = {k: v for k, v in vars(user_context.login_org).items()}

        data = json.dumps(
            {
                "permissions": permissions,
                "entities": entities,
                "caller": caller,
                "env": env or {}
            })
        headers = dict(request.headers)
        headers['Content-Type'] = 'application/json'
        method = "POST"
        url = '{gateway_url}/access/v1/demand'.format(gateway_url=self.helper_service.get_gateway_url(gateway_ip))
        self.app.logger.debug(f"Request {method} {url}")
        req_start = time.time()
        rbac_response = requests.request(
            method=method,
            url=url,
            data=data,
            headers=headers
        )
        req_rtt = time.time() - req_start
        self.app.logger.debug(f"Response status code {rbac_response.status_code}; Took {req_rtt:.4f}s")

        if not rbac_response.ok:
            raise Forbidden

        return rbac_response
