import os
import time

import ipaddress
import requests
import simplejson as json
import re
import uuid
import hashlib

from models.call_context import CallContext
from models.request_context import RequestContext
from models.session_context import SessionContext
from models.user_context import UserContext
from services.cache_service import CacheService
from services.connector_service import ConnectorLookupService
from services.helper_service import HelperService
from werkzeug.exceptions import Forbidden
from models.custom_exceptions.auth_token_parse_exception import AuthTokenParseException


class AuthContextService(object):
    gateway_host = os.getenv('GATEWAY_HOST')
    gateway_port = os.getenv('GATEWAY_PORT')
    gateway_csr_port = os.getenv('GATEWAY_CSR_PORT')
    gateway_authed_port = os.getenv('GATEWAY_AUTHED_PORT')
    trusted_proxies = os.getenv('TRUSTED_PROXIES', '10.0.0.0/8')
    api_key_salt = 'B3b11-!$1NLfzk0$s'

    def __init__(self, app, xcsrf_service, helper_service=None, connector_service=None, cache_service=None):
        self.app = app
        self.xcsrf_service = xcsrf_service
        self.helper_service = helper_service or HelperService()
        self.connector_service = connector_service or ConnectorLookupService(app=self.app)
        self.cache_service = cache_service or CacheService(app)
        self.set_trusted_proxies(AuthContextService.trusted_proxies)
        self.connector_key_timeout = int(float(os.getenv('CONNECTOR_KEY_TIMEOUT', 15)) * 60)

    def set_trusted_proxies(self, proxies):
        trusted = [x for x in proxies.split(',') if len(x) > 0]
        self.trusted_networks = [ipaddress.IPv4Network(x.strip()) for x in trusted]

    def get_request_context(self, request):
        if not request.headers.get('X-Request-Context'):
            return self._build_request_context(request)

        return RequestContext.parse(json.loads(request.headers.get('X-Request-Context')))

    def get_call_context(self, request):
        ctx = CallContext()
        ctx.auth_token = self._get_auth_token(request)
        ctx.session_id = self._get_session_id(request)
        return ctx

    def create_call_context(self, request):
        ctx = CallContext()
        ctx.auth_token = self._get_auth_token(request)
        if not ctx.auth_token:
            # Only generate a session ID when there is no auth token
            ctx.session_id = self._create_session_id(request)
        return ctx

    def get_session_context(self, request, request_ctx=None):
        request_ctx = request_ctx or self.get_request_context(request)
        call_ctx = self.get_call_context(request)
        user_ctx = UserContext()

        if self._is_whitelisted_request(request):
            call_ctx.authenticated = True
            return SessionContext(call_context=call_ctx, authenticated=True, origin_env=request_ctx.origin_env)

        session_id = None
        if call_ctx.auth_token:
            session_id = self.get_session_id_from_auth_token(call_ctx.auth_token)
        else:
            session_id = call_ctx.session_id

        session_extension_required = self._session_extension_required(request)

        session_ctx = self.cache_service.get_session_context(session_id, session_extension_required)

        valid_session_context = False
        if session_ctx:
            if not session_ctx.origin_env:
                self.app.logger.warning("Session '{}' does not have an 'origin_env' set".format(session_id))
                session_ctx.origin_env = request_ctx.origin_env
                self.cache_service.set_session_context(session_id, session_ctx)
                valid_session_context = True
            elif session_ctx.origin_env == request_ctx.origin_env:
                # self.app.logger.debug("Session '{}' matched request 'origin_env'".format(session_id))
                valid_session_context = True
            else:
                self.app.logger.warning("Session '{}' mismatch request 'origin_env={}'".format(session_id,
                                                                                               request_ctx.origin_env))

        if valid_session_context:
            # Extend the session if it is not a connector and session extension is required
            if call_ctx.is_session_call() and session_extension_required:
                self.cache_service.extend_session(session_ctx.auth_id, session_id)
            if not HelperService.is_csrf_exempt_authenticated_request(request):
                self.xcsrf_service.validate_token(session_ctx.xcsrf)
            return session_ctx

        ############## Creating session for current call
        session_ctx = SessionContext(
            call_context=call_ctx,
            user_context=user_ctx,
            origin_env=request_ctx.origin_env
        )

        if call_ctx.auth_token:
            ############## If this call is using auth_token and there is no session this must be a first call
            ############## We will populate the session for it
            session_ctx = self._handle_connector_auth(request, request_ctx, session_ctx, session_id, call_ctx, user_ctx)

        return session_ctx

    def _handle_connector_auth(self, request, request_ctx, session_ctx, session_id, call_ctx, user_ctx):
        connector = self.connector_service.validate_auth_token(request_ctx, call_ctx.auth_token, request.headers)
        if not connector:
            session_ctx = SessionContext(call_context=call_ctx, authenticated=False, origin_env=request_ctx.origin_env)
        else:
            try:
                session_ctx = self.connector_service.process_connector_login(
                    request,
                    request_ctx,
                    call_ctx,
                    user_ctx,
                    connector
                )
                self.cache_service.set_session_context(session_id, session_ctx, self.connector_key_timeout)
            except Forbidden:
                # Logging is done on the access service level when the RBAC key cannot be found for a connector
                pass

        return session_ctx

    def make_response(self, request, request_ctx, session_ctx):
        ui_message = None
        message_json = session_ctx.user_info if session_ctx.authenticated else {}
        message_json["success"] = session_ctx.authenticated
        message_json["message"] = "User is {0}authenticated".format('' if session_ctx.authenticated else 'not ')
        message = json.dumps(message_json)

        if not session_ctx.authenticated:
            ui_message = self.is_ui_url(request, "{gateway_url}" + self._get_requested_url(request))
            message = ui_message or message

        resp_code = 401 if not session_ctx.authenticated else 200
        response = self.app.make_response((message, resp_code))

        if not session_ctx.authenticated:
            response.delete_cookie("SESSION", path='/')
            response.delete_cookie("SESSION", path='/appservices/')
            response.delete_cookie("SESSION", path='/adminServices/')

        response.headers['X-Processed-By'] = "psc.auth"
        response.headers['Content-Security-Policy'] = "frame-ancestors 'self';"
        response.headers['Strict-Transport-Security'] = "max-age=63072000; includeSubDomains; preload"
        response.headers['Referrer-Policy'] = "origin"

        if not ui_message:
            response.headers['Content-Type'] = "application/json; charset=utf-8"

        if session_ctx.authenticated and not request.headers.get("X-Request-Context"):
            response.headers["X-Request-Context"] = request_ctx.json()
            response.headers["X-User-Context"] = session_ctx.user_context.json()
            response.headers["X-OrgKey"] = session_ctx.user_context.target_org.org_key
            ip_address = str(ipaddress.IPv6Address(session_ctx.user_context.rate_org_id()))
            response.headers["X-Org-Rate-Ip"] = ("0" + ip_address) if (ip_address.startswith("::")) else ip_address

        # self.app.logger.debug("Make response for session: '{}'; body: '{}'".format(
        #                        session_ctx.call_context.session_id,
        #                        message))
        return response

    def get_session_id_from_auth_token(self, auth_token):
        try:
            connector = self.connector_service.parse_auth_token(auth_token)
            h = hashlib.md5()
            h.update(bytes(f'{connector.api_key}{self.api_key_salt}', 'utf-8'))
            session_id = f'{h.hexdigest()}/{connector.connector_id}'
        except AuthTokenParseException:
            session_id = None

        return session_id

    def _get_gateway_ip(self, request):
        forwarded_ips = request.headers.get('X-Forwarded-For', '').split(',')
        forwarded_uri = request.headers.get('X-Forwarded-Uri')
        gateway_dns = 'gateway.psc.internal'

        if forwarded_uri:
            return forwarded_ips[-1].strip() if len(forwarded_ips) > 1 else self.gateway_host

        return gateway_dns

    def _is_trusted_ip(self, ip):
        for network in self.trusted_networks:
            if ip in network:
                return True
        return False

    def _get_client_ip(self, request):
        x_forwarded = request.headers.get('X-Forwarded-For', '')

        forwarded_ips = []
        for ip_str in x_forwarded.split(','):
            try:
                ip = ipaddress.IPv4Address(ip_str.strip())
                forwarded_ips.append(ip)
            except ValueError:
                self.app.logger.warn(f'Unable to parse IPv4 address: {ip_str}')

        forwarded_ips.reverse()

        for ip in forwarded_ips:
            if not self._is_trusted_ip(ip):
                return str(ip)

    def _build_request_context(self, request):
        ctx = RequestContext()
        ctx.gateway_ip = self._get_gateway_ip(request)
        ctx.gateway_port = self.gateway_port
        ctx.gateway_csr_port = self.gateway_csr_port
        ctx.gateway_authenticated_port = self.gateway_authed_port
        ctx.gateway_public_url = request.headers["X-Forwarded-Host"]
        ctx.set_origin_env(request.headers["X-Forwarded-Host"])
        ctx.client_ip = self._get_client_ip(request)
        return ctx

    def _get_auth_token(self, request):
        return request.headers.get('X-Auth-Token')

    def _get_session_id(self, request):
        return request.cookies.get('SESSION')

    def _create_session_id(self, request):
        # Todo: Remove the request method param, and do not check session cookie once the backend has fully disabled
        # session creation
        session_id = request.cookies.get('SESSION')
        if session_id is None:
            session_id = str(uuid.uuid4())
            self.app.logger.info("New session ID created: {session_id}".format(session_id=session_id))
        return session_id

    def _get_requested_url(self, request):
        forwarded_uri = request.headers.get('X-Forwarded-Uri')
        return forwarded_uri or request.url_rule.rule

    def _get_original_method(self, request):
        forwarded_method = request.headers.get('X-Forwarded-Method')
        return forwarded_method or request.method

    def _is_whitelisted_request(self, request):
        url = self._get_requested_url(request)
        method = self._get_original_method(request)

        # Tuple with a regex for a whitelisted URL, and whitelisted HTTP methods.  All methods whitelisted if empty.
        whitelist = [
            (r'^/$', []),
            (r'^/\?.*', []),
            (r'^/resetpassword.*', []),
            (r'^/adminconfirm.*', []),
            (r'^/checkAuthStrategy.*', []),
            (r'^/logout.*', []),
            (r'^/login.*', []),
            (r'^/samlLogin.*', []),
            (r'^/favicon\.ico.*', []),
            (r'^/assets.*', []),
            (r'^/appservices/v5/admin/login/.*', []),
            (r'^/[a-z0-9]+_bundle\.[a-z0-9]+\.js*', []),
            (r'^/appservices/v5/user/.*/code', []),
            (r'^/appservices/v5/user/.*/password', ['POST', 'GET']),
            (r'^/auth/v1/sessions', ['POST']),
            (r'^/auth/v1/sessions/saml', ['POST'])
        ]

        for tuple in whitelist:
            url_pattern = tuple[0]
            methods = tuple[1]

            if re.match(url_pattern, url) and (len(methods) == 0 or method in methods):
                self.app.logger.info("Request is whitelisted: {}:{}".format(method, url))
                return True
        return False

    def _session_extension_required(self, request):
        """
        Determines if a session extension is required for a request based on the X-Forwarded-URI in the request object
        :param request
        :return: true if the request is not in the list of excluded paths
        """
        url = self._get_requested_url(request)
        method = self._get_original_method(request)

        # Tuple with a regex for a excluded URLs for session extension, and HTTP methods.
        # All methods excluded if empty.
        excluded_paths = [
            (r'^/notification/v1/notifications', ['GET'])
        ]

        for tup in excluded_paths:
            url_pattern = tup[0]
            methods = tup[1]
            if re.match(url_pattern, url) and (len(methods) == 0 or method in methods):
                return False
        return True

    def is_ui_url(self, request, url):
        try:
            mime_type = request.accept_mimetypes.best_match(['text/html'])
            method = request.method
            call_ctx = self.get_call_context(request)
            if mime_type == "text/html" and method == "GET" and not call_ctx.is_api_call():
                req_ctx = self.get_request_context(request)
                gateway_url = self.helper_service.get_gateway_url(
                    gateway_ip=req_ctx.gateway_ip,
                    port=req_ctx.gateway_port
                )
                headers = self.helper_service.get_headers_copy(request.headers, 'If-Modified-Since', 'If-None-Match')
                headers['Host'] = headers.get('X-Forwarded-Host')
                url = url.format(gateway_url=gateway_url)
                self.app.logger.debug(f"Request {request.method} {url}")
                req_start = time.time()
                response = requests.request(
                    method=request.method,
                    url=url,
                    headers=headers
                )
                req_rtt = time.time() - req_start
                self.app.logger.debug(f"Response status code {response.status_code}; Took {req_rtt:.4f}s")
                is_response_from_ui_server = response.headers.get('X-Processed-By', "") == 'cbcui' \
                    or response.headers.get('Server', "").startswith('nginx')

                if response.status_code == 200 \
                        and response.headers.get('Content-Type', "text/html") == "text/html" \
                        and is_response_from_ui_server:
                    return response.text

                if response.status_code == 304 \
                        and is_response_from_ui_server:
                    return response.text

        except:
            pass

        return None
