import os
from datetime import datetime
from redis.connection import Connection
import simplejson as json
import re

from flask import Flask
from werkzeug.exceptions import NotFound

from models.session_context import SessionContext

_app = Flask(__name__)

pre_patched = Connection.connect


def patched_connect(self):
    re_open = True
    try:
        re_open = (datetime.now() - self._opened_at).seconds > 300
        if re_open:
            self.disconnect()
            _app.logger.debug("Re-opening the connection")
        else:
            _app.logger.debug("Re-using the connection")
    except AttributeError:
        self._opened_at = datetime.now()

    pre_patched(self)

    if re_open:
        self._opened_at = datetime.now()


Connection.connect = patched_connect


class CacheService(object):
    redis_url = os.getenv('REDIS_URL')
    timeout = int(float(os.getenv('SESSION_TIMEOUT', 65)) * 60)

    def __init__(self, app):
        import redis
        self.app = app
        self.r = redis.StrictRedis(host=self.redis_url, port=6379, db=0, socket_timeout=5, socket_connect_timeout=2,
                                   socket_keepalive=True, retry_on_timeout=True)

    def get_session_cache_key(self, session_id):
        return 'session:{session_id}'.format(session_id=session_id)

    def get_auth_id_cache_key(self, session_id, auth_id):
        return 'session:{session_id}:auth_id:{auth_id}'.format(session_id=session_id, auth_id=auth_id)

    def get_rbac_cache_key(self, auth_id):
        return 'session:auth_id:{auth_id}:rbac_key'.format(auth_id=auth_id)

    def get_auth_id_cache_pattern(self, auth_id='*', session_id='*'):
        return 'session:{session_id}:auth_id:{auth_id}'.format(session_id=session_id, auth_id=auth_id)

    def parse_auth_id_cache_key(self, cache_key):
        if not re.match("session:.*:auth_id:.*", cache_key):
            return None
        parts = re.split(":", cache_key)
        return parts[1], parts[3]

    def extend(self, key):
        if not key:
            return
        self.r.expire(name=key, time=self.timeout)

    def extend_session(self, auth_id, session_id):
        auth_id_cache_key = self.get_auth_id_cache_key(session_id, auth_id)
        session_cache_key = self.get_session_cache_key(session_id)
        rbac_cache_key = self.get_rbac_cache_key(auth_id)
        self.extend(auth_id_cache_key)
        self.extend(session_cache_key)
        self.extend(rbac_cache_key)

    """
        Matches a session context cache to an orgkey. Needed because there could be a session cache from a different
        environment that has an identical auth id. This ensures that we don't blow away a user's session if they're in
        another environment but happen to have the same auth id.
        :return True if the session context cache login org_key or target org_key matches the given org_key
    """
    def matching_session_context(self, session_id, org_key):
        if not session_id:
            return False
        session_context = self.r.get(self.get_session_cache_key(session_id))

        if session_context:
            session_context = SessionContext.parse(json.loads(session_context.decode('utf-8')))

            session_ctx_login_org_key = session_context.user_context.login_org.org_key
            session_ctx_target_org_key = session_context.user_context.target_org.org_key

            if org_key == session_ctx_login_org_key or org_key == session_ctx_target_org_key:
                return True

        return False

    def get_session_context(self, session_id, extend_session=True):
        if not session_id:
            return None
        session_context = self.r.get(self.get_session_cache_key(session_id))

        if session_context:
            session_context = SessionContext.parse(json.loads(session_context.decode('utf-8')))
            if extend_session:
                self.extend_session(session_context.auth_id, session_id)

            # Update the rbac key
            rbac_key = self.get_session_rbac_key(session_context.auth_id)
            session_context.user_context.target_org.rbac_key = rbac_key
            session_context.user_context.login_org.rbac_key = rbac_key

            # self.app.logger.debug("Got session context for session_id='{}' from cache: \
            #                       '{}'".format(session_id, session_context.json()))
        # else:
            # self.app.logger.debug("No session context for session_id='{}' in cache".format(session_id))

        return session_context

    def set_session_context(self, session_id, session_context, timeout=None):
        if not session_id:
            raise NotFound("session_id not found")

        rbac_key = session_context.user_context.login_org.rbac_key
        if not rbac_key:
            self.app.logger.warning("Setting session context with an empty rbac key for session: {}".format(session_id))

        self.r.set(self.get_auth_id_cache_key(session_id, session_context.auth_id), 1, ex=timeout or self.timeout)
        self.r.set(self.get_rbac_cache_key(session_context.auth_id), rbac_key, ex=timeout or self.timeout)
        # self.app.logger.debug("Setting cache key '{}' to: '{}'".format(self.get_session_cache_key(session_id),
        #                                                               session_context.json()))
        return self.r.set(self.get_session_cache_key(session_id), session_context.json(), ex=timeout or self.timeout)

    def purge_user_cache(self, session_id, auth_id):
        if not session_id:
            return

        self.r.delete(self.get_auth_id_cache_key(session_id, auth_id))

        self.r.delete(
            self.get_session_cache_key(session_id),
            'spring:session:sessions:{session_id}'.format(session_id=session_id)
        )

    # Delete all sessions belonging to the auth ID
    def evict_sessions(self, auth_id):
        if not auth_id:
            return 0

        # Delete the entity's RBAC key cache entry
        self.r.delete(self.get_rbac_cache_key(auth_id))

        # Delete each session context and auth ID cache entry
        key_pattern = self.get_auth_id_cache_pattern(auth_id=auth_id)
        evicted = 0
        for key in self.r.scan_iter(key_pattern):
            key = key.decode("utf-8")
            del_key = self.r.delete(key)
            session_id, auth_id = self.parse_auth_id_cache_key(key)
            session_cache_key = self.r.delete(self.get_session_cache_key(session_id))

            if del_key and session_cache_key:
                evicted += 1

        return evicted

    # Delete all user sessions except the current one
    def reset_sessions(self, auth_id, org_key, current_session_id):
        if not auth_id or not org_key or not current_session_id:
            return 0
        key_pattern = self.get_auth_id_cache_pattern(auth_id=auth_id)
        num_deleted = 0
        for key in self.r.scan_iter(key_pattern):
            key = key.decode("utf-8")
            session_id, auth_id = self.parse_auth_id_cache_key(key)
            is_matching_session = self.matching_session_context(session_id, org_key)
            if is_matching_session and session_id != current_session_id:
                self.r.delete(key)
                self.r.delete(self.get_session_cache_key(session_id))
                num_deleted += 1

        return num_deleted

    # Set the rbac key cache entry only if there is an existing session matching the auth id
    def update_session_rbac_key(self, auth_id, claims, timeout=None):
        key_pattern = self.get_auth_id_cache_pattern(auth_id=auth_id)

        session_exists = len([key for key in self.r.scan_iter(key_pattern)]) > 0

        if auth_id and session_exists:
            self.set_session_rbac_key(auth_id, claims, timeout or self.timeout)

    def set_session_rbac_key(self, auth_id, claims, timeout=None):
        if auth_id:
            self.r.set(self.get_rbac_cache_key(auth_id), claims, ex=timeout or self.timeout)

    def get_session_rbac_key(self, auth_id):
        if auth_id:
            rbac_key = self.r.get(self.get_rbac_cache_key(auth_id))
            if not rbac_key:
                self.app.logger.warning("No RBAC key found in cache for entity: {}".format(auth_id))
            return rbac_key
