import json
import os
import time

import requests

from models.connector import Connector
from models.session_context import SessionContext

from services.access_service import AccessService
from services.helper_service import HelperService

from models.custom_exceptions.connector_not_authenticated_exception import ConnectorNotAuthenticatedException
from models.custom_exceptions.auth_token_parse_exception import AuthTokenParseException


class ConnectorLookupService(object):
    def __init__(self, helper_service=None, access_service=None, app=None):
        self.helper_service = helper_service or HelperService()
        self.access_service = access_service or AccessService(app=app)
        self.app = app
        self.request_timeout = int(float(os.getenv("REQUEST_TIMEOUT", 25)))

    def parse_auth_token(self, x_auth_token):
        connector = Connector()
        connector.auth_token = x_auth_token

        keys = x_auth_token.split('/')
        if len(keys) < 2:
            raise AuthTokenParseException

        connector.api_key = keys[0]
        connector.connector_id = keys[1]
        connector.api_session = keys[2] if len(keys) == 3 else None
        return connector

    def validate_auth_token(self, request_ctx, x_auth_token, headers):
        connector = None
        try:
            login_headers = {}
            x_forwarded_for = headers.get("X-Forwarded-For", None)
            # If an x-forwarded-for header exists and the client_IP is not a trusted IP
            if request_ctx.client_ip is not None and x_forwarded_for is not None:
                login_headers["X-Forwarded-For"] = request_ctx.client_ip + (", " + x_forwarded_for)
            # If an x-forwarded-for header exists but the client_IP was trusted so it is null
            elif request_ctx.client_ip is None and x_forwarded_for is not None:
                login_headers["X-Forwarded-For"] = x_forwarded_for
            # If the x-forwarded-for header value doesn't exist at all
            else:
                login_headers["X-Forwarded-For"] = ""

            login_headers["X-Request-Context"] = request_ctx.json()
            login_headers["X-Auth-Token"] = headers.get("X-Auth-Token", None)
            login_headers["Host"] = headers.get("X-Forwarded-Host", None)
            login_headers["Content-Type"] = "application/json"

            url = '{gateway_url}/appservices/v5/connector/login'
            url = url.format(gateway_url=self.helper_service.get_gateway_url(request_ctx.gateway_ip))
            method = "POST"
            self.app.logger.debug(f"Request {method} {url}")
            req_start = time.time()
            _response = requests.request(
                method=method,
                url=url,
                headers=login_headers,
                timeout=self.request_timeout
            )
            req_rtt = time.time() - req_start
            self.app.logger.debug(f"Response status code {_response.status_code}; Took {req_rtt:.4f}s")
            connector = self.parse_auth_token(x_auth_token)

            if _response.status_code != 200:
                self.app.logger.warning("Appservices call to authenticate connector {} has failed. "
                                        "Status Code {}. "
                                        "Response {} "
                                        .format(connector.connector_id, _response.status_code, _response.text))
                raise ConnectorNotAuthenticatedException

            connector_entry = json.loads(_response.text)
            connector.org_id = connector_entry["currentOrgId"]
            connector.org_key = connector_entry["currentOrgKey"]
            connector.connector_type = connector_entry["connectorType"]
            return connector
        except AuthTokenParseException:
            self.app.logger.warning("Auth-Token {} is not properly constructed".format(x_auth_token))
        except ConnectorNotAuthenticatedException:
            pass

        return None

    def process_connector_login(self, request, request_ctx, call_ctx, user_ctx, connector):
        session_ctx = SessionContext(
            call_context=call_ctx,
            user_context=user_ctx,
            origin_env=request_ctx.origin_env
        )
        call_ctx.api_key = connector.api_key
        call_ctx.api_session = connector.api_session
        call_ctx.connector_id = connector.connector_id
        call_ctx.connector_type = connector.connector_type
        user_ctx.login_org.org_key = connector.org_key
        user_ctx.login_org.org_id = connector.org_id

        user_ctx.login_org.connector_id = connector.connector_id
        user_ctx.login_org.rbac_key = self.access_service.get_session_rbac_key(
            request_ctx.gateway_ip,
            request.headers,
            user_ctx.login_org.org_key,
            None,
            call_ctx.connector_id,
            call_ctx.connector_type
        )
        user_ctx.target_org = user_ctx.login_org
        session_ctx.authenticated = True
        session_ctx.auth_id = connector.connector_id
        return session_ctx
