import os
import re
import time

import simplejson as json
import requests
from werkzeug.exceptions import Forbidden
from werkzeug.utils import redirect

from services.auth_ctx_service import AuthContextService
from services.helper_service import HelperService
from services.access_service import AccessService
from services.cache_service import CacheService
from services.connector_service import ConnectorLookupService

from models.user_context import UserContext
from models.session_context import SessionContext

from services.xcsrf_service import CSRF_TOKEN_HEADER

from models.context_headers_response import ContextHeaders


def _get_login_status(response):
    response_body = json.loads(response.text)

    return response_body['loginStatus']


class LoginService(object):
    secure_cookie = os.getenv('SECURE_COOKIE', 'false').lower() == "true"
    ALLOWED_MDR_ROLES_FOR_GET_CONTEXT_API = ["MDR_USER_CONTEXT_SERVICE_ROLE"]

    def __init__(self, app, xsrf_service, auth_ctx_service=None, helper_service=None, connector_service=None,
                 cache_service=None, access_service=None):
        self.app = app
        self.xsrf_service = xsrf_service
        self.auth_ctx_service = auth_ctx_service or AuthContextService(self.app, self.xsrf_service)
        self.helper_service = helper_service or HelperService()
        self.connector_service = connector_service or ConnectorLookupService(app=self.app)
        self.cache_service = cache_service or CacheService(self.app)
        self.access_service = access_service or AccessService(app=self.app)
        self.request_timeout = int(float(os.getenv("REQUEST_TIMEOUT", 25)))

    def _make_response(self, response, redirect_url=None):
        if redirect_url:
            return redirect(redirect_url)
        else:
            response = self.app.make_response((response.text, response.status_code))
            response.headers['X-Processed-By'] = 'psc.auth'
            response.headers['Content-Type'] = 'application/json; charset=utf-8'
            return response

    def _make_response_from_message(self, message, status_code=200):
        response = self.app.make_response((json.dumps({
            "success": status_code == 200,
            "message": message
        }), status_code))
        response.headers['X-Processed-By'] = 'psc.auth'
        response.headers['Content-Type'] = 'application/json; charset=utf-8'
        return response

    def _make_response_from_content(self, content, status_code=200):
        response = self.app.make_response((content, status_code))
        response.headers['X-Processed-By'] = 'psc.auth'
        response.headers['Content-Type'] = 'application/json; charset=utf-8'
        return response

    def login(self, request, url, data, saml=False):
        headers = self.helper_service.get_headers_copy(request.headers, 'Content-Length', 'Cookie', 'Accept-Encoding')
        headers['Accept-Encoding'] = 'gzip;q=0,deflate;q=0'
        headers['Content-Type'] = 'application/json'
        req_ctx = self.auth_ctx_service.get_request_context(request)
        gateway_url = self.helper_service.get_gateway_url(
            gateway_ip=req_ctx.gateway_ip,
            port=req_ctx.gateway_port
        )

        url = url.format(gateway_url=gateway_url)
        self.app.logger.debug(f"Request {request.method} {url}")
        req_start = time.time()
        response = requests.request(
            method=request.method,
            url=url,
            data=data,
            headers=headers,
            timeout=self.request_timeout
        )
        req_rtt = time.time() - req_start
        self.app.logger.debug(f"Response status code {response.status_code}; Took {req_rtt:.4f}s")

        if response.status_code == 200:
            redirect_url = None
            session_ctx = None
            error_param = None

            if saml:
                redirect_url = request.headers['X-Forwarded-Proto'] + "://" + req_ctx.gateway_public_url
                error_param = '?error_code=' + _get_login_status(response)

            try:
                session_ctx = self._process_login_success(request, response, req_ctx)
            except Forbidden as e:
                # Logging is done on the access service level
                return self._make_response_from_message(e.description, e.code)

            if not session_ctx:
                if redirect_url is not None and error_param is not None:
                    redirect_url = redirect_url + error_param
                return self._make_response(response, redirect_url)

            call_ctx = session_ctx.call_context
            self.cache_service.set_session_context(call_ctx.session_id, session_ctx)
            response = self._make_response(response, redirect_url)
            response.headers[CSRF_TOKEN_HEADER] = session_ctx.xcsrf
            response.set_cookie(
                "SESSION",
                call_ctx.session_id,
                secure=self.secure_cookie,
                httponly=True,
            )
        else:
            mes = 'Error login response. Actual: {}: {}'.format(response.status_code, response.text)
            self.app.logger.error(mes)
            response = self._make_response(response)
            response.delete_cookie("SESSION", path='/')
            response.delete_cookie("SESSION", path='/appservices/')
            response.delete_cookie("SESSION", path='/adminServices/')
            response.status_code = 200

        # workaround for chunked returning a incomplete parsing error
        response.headers = self.helper_service.remove_chunked_header(response.headers)

        return response

    def switch_org(self, request, org_id, url):
        headers = self.helper_service.get_headers_copy(request.headers, 'Content-Length', 'Accept-Encoding')
        headers['Accept-Encoding'] = 'gzip;q=0,deflate;q=0'

        req_ctx = self.auth_ctx_service.get_request_context(request)

        gateway_url = self.helper_service.get_gateway_url(
            gateway_ip=req_ctx.gateway_ip,
            port=req_ctx.gateway_port
        )

        url = url.format(gateway_url=gateway_url, org_id=org_id)
        method = "POST"
        self.app.logger.debug(f"Request {method} {url}")
        req_start = time.time()
        response = requests.request(
            method=method,
            url=url,
            data=request.data,
            headers=headers
        )
        req_rtt = time.time() - req_start
        self.app.logger.debug(f"Response status code {response.status_code}; Took {req_rtt:.4f}s")

        response.headers['X-Processed-By'] = "psc.auth"

        if response.status_code == 200:
            session_ctx = self._process_success_org_switch(request, response, req_ctx)
            self.cache_service.set_session_context(session_ctx.call_context.session_id, session_ctx)
        else:
            mess = "Org call failed for {} with code {}, text {}".format(url, response.status_code, response.text)
            self.app.logger.error(mess)

        # workaround for chunked returning a incomplete parsing error
        response.headers = self.helper_service.remove_chunked_header(response.headers)
        return response.text, response.status_code, response.headers.items()

    def _process_login_success(self, request, response, req_ctx):
        user_info = json.loads(response.text)
        if not user_info.get('success', None):
            return None

        user_info["loginOrgId"] = user_info["currentOrgId"]
        user_info["loginOrgKey"] = user_info["currentOrgKey"]

        user_ctx = UserContext()
        call_ctx = self.auth_ctx_service.create_call_context(response)
        user_ctx.login_org.user_id = user_info["loginId"] if user_info else None
        user_ctx.login_org.user_name = user_info["loginName"] if user_info else None
        user_ctx.login_org.org_id = user_info["currentOrgId"] if user_info else None
        user_ctx.login_org.org_key = user_info["currentOrgKey"] if user_info else None
        user_ctx.target_org = user_ctx.login_org

        headers = dict(request.headers)
        headers.pop('Content-Length', None)

        headers["X-User-Context"] = user_ctx.json()

        user_ctx.login_org.rbac_key = self.access_service.get_session_rbac_key(
            req_ctx.gateway_ip,
            headers,
            user_ctx.login_org.org_key,
            user_ctx.login_org.user_id,
            None,
            None
        )

        user_ctx.target_org = user_ctx.login_org
        return SessionContext(
            call_context=call_ctx,
            user_context=user_ctx,
            user_info=user_info,
            authenticated=True,
            xcsrf=self.xsrf_service.generate_token(),
            client_ip=req_ctx.client_ip,
            auth_id=user_info["loginId"] if user_info else None,
            origin_env=req_ctx.origin_env
        )

    def _process_success_org_switch(self, request, response, req_ctx):
        try:
            org_response = json.loads(response.text)
        except ValueError:
            message = 'Error parsing json from response. Actual: {}: {}'.format(response.status_code, response.content)
            self.app.logger.error(message)
            raise ValueError("Error parsing json from response")

        organization = org_response["organization"]
        session_ctx = self.auth_ctx_service.get_session_context(request, req_ctx)

        user_ctx = session_ctx.user_context
        user_ctx.target_org.org_id = int(organization['id'])
        user_ctx.target_org.org_key = organization["orgKey"]

        user_info = session_ctx.user_info
        user_info['currentOrgId'] = user_ctx.target_org.org_id
        user_info['currentOrgKey'] = user_ctx.target_org.org_key

        return session_ctx

    def verify_user_email(self, request, email, url):
        headers = self.helper_service.get_headers_copy(request.headers, 'Content-Length', 'Cookie')
        headers['Content-Type'] = 'application/json'

        # Uncomment below line for any debugging
        # self.app.logger.warning("Get Context API: Request headers {}".format(headers))

        req_ctx = self.auth_ctx_service.get_request_context(request)

        # If user's role is not from the predefined ALLOWED_LIST, throw a 502
        session_ctx = self.auth_ctx_service.get_session_context(request, req_ctx)
        if session_ctx and session_ctx.user_context:
            rbac_key_str = session_ctx.user_context.target_org.rbac_key.decode('utf-8')

            # checking if string contains list element
            rbac_key_contains_role = [role for role in LoginService.ALLOWED_MDR_ROLES_FOR_GET_CONTEXT_API
                                      if (role in rbac_key_str)]
            if not rbac_key_contains_role:
                return self._make_response_from_message(message="Bad Gateway", status_code=502)

        gateway_url = self.helper_service.get_gateway_url(
            gateway_ip=req_ctx.gateway_ip,
            port=req_ctx.gateway_csr_port  # call to adminServices doesn't work on internal port(8902)
        )

        url = url.format(gateway_url=gateway_url, email=email)
        self.app.logger.debug(f"Request {request.method} {url}")
        req_start = time.time()
        response = requests.request(
            method=request.method,
            url=url,
            headers=headers,
            timeout=self.request_timeout
        )
        req_rtt = time.time() - req_start
        self.app.logger.debug(f"Response status code {response.status_code}; Took {req_rtt:.4f}s")

        # Set to warning for dev01 debugging
        self.app.logger.info("Get Context API: Status code {} from adminServices/v5/users/{}/, Response {}"
                             .format(response.status_code, email, response.text))

        if response.status_code == 200:
            try:
                context_headers = self._process_login_details_success(request, response, req_ctx)
                response = self._make_response_from_content(context_headers.json(), response.status_code)
            except Forbidden as e:
                response = self._make_response_from_message(e.description, e.code)
        else:
            self.app.logger.warning("AdminService call failed for adminServices/v5/users/{}/ with status_code {}, "
                                    "Response {}".format(email, response.status_code, response.text))
            response = self._make_response(response)

        # workaround for chunked returning a incomplete parsing error
        response.headers = self.helper_service.remove_chunked_header(response.headers)

        return response

    def _process_login_details_success(self, request, response, req_ctx):
        try:
            user_details_from_response = json.loads(response.text)
        except ValueError:
            message = 'Error parsing json from response. Actual: {}: {}'.format(response.status_code, response.content)
            self.app.logger.error(message)

        if not user_details_from_response.get('success', None):
            return None

        user_info = user_details_from_response.get('user', None)
        user_ctx = UserContext()
        user_ctx.login_org.user_id = user_info["loginId"] if user_info else None
        user_ctx.login_org.user_name = user_info["loginName"] if user_info else None
        user_ctx.login_org.org_id = user_info["orgId"] if user_info else None
        user_ctx.login_org.org_key = user_info["orgKey"] if user_info else None
        user_ctx.target_org = user_ctx.login_org

        headers = dict(request.headers)
        headers.pop('Content-Length', None)

        user_ctx.login_org.rbac_key = self.access_service.get_session_rbac_key(
            req_ctx.gateway_ip,
            headers,
            user_ctx.login_org.org_key,
            user_ctx.login_org.user_id,
            None,
            None
        )

        user_ctx.target_org = user_ctx.login_org
        return ContextHeaders(
            user_context=user_ctx,
            request_context=req_ctx,
        )
