from flask import (_app_ctx_stack, current_app, has_request_context, request)
from werkzeug.exceptions import Forbidden
from werkzeug.security import safe_str_cmp

from services.flask_seasurf.flask_seasurf import REASON_NO_REFERER, \
    REASON_BAD_REFERER, \
    REASON_BAD_TOKEN, \
    REASON_NO_REQUEST, \
    _same_origin, \
    SeaSurf

CSRF_TOKEN_HEADER = 'X-CSRF-Token'


class XCSRFService(SeaSurf):
    def __init__(self, app):
        app.config['CSRF_HEADER_NAME'] = CSRF_TOKEN_HEADER
        super(XCSRFService, self).__init__(app)
        self._disable_cookie = lambda _: True

    def generate_token(self):
        return self._generate_token()

    def _should_use_token(self, view_func):
        if "X-Auth-Token" in request.headers and self._csrf_header_name not in request.headers:
            return False
        return super(XCSRFService, self)._should_use_token(view_func)

    def validate_token(self, token):
        self.xsrf_token = token
        self.__before_request()
        self.xsrf_token = None

    def __before_request(self):
        # Always set this to let the response know whether or not to set the
        # CSRF token.
        _app_ctx_stack.top._view_func = \
            current_app.view_functions.get(request.endpoint)

        if request.headers.get('X-Forwarded-Method') \
                and request.headers.get('X-Forwarded-Method') not in ('GET', 'HEAD', 'OPTIONS', 'TRACE'):
            # Retrieve the view function based on the request endpoint and
            # then compare it to the set of exempted views
            if not self._should_use_token(_app_ctx_stack.top._view_func):
                return

            self.validate()
        return

    def _before_request(self):
        return

    def validate(self):
        if not has_request_context():
            raise Forbidden(description=REASON_NO_REQUEST)

        # Get this from our auth session only if it's already authenticated
        server_csrf_token = self.xsrf_token

        if request.is_secure and self._check_referer:
            referer = request.headers.get('Referer')
            if referer is None:
                error = (REASON_NO_REFERER, request.path)
                error = u'Forbidden ({0}): {1}'.format(*error)
                current_app.logger.warning(error)
                raise Forbidden(description=REASON_NO_REFERER)

            # By setting the Access-Control-Allow-Origin header, browsers
            # will let you send cross-domain AJAX requests so if there is
            # an Origin header, the browser has already decided that it
            # trusts this domain otherwise it would have blocked the
            # request before it got here.
            allowed_referer = request.headers.get('Origin') or request.url_root
            if not _same_origin(referer, allowed_referer):
                error = REASON_BAD_REFERER.format(referer, allowed_referer)
                description = error
                error = (error, request.path)
                error = u'Forbidden ({0}): {1}'.format(*error)
                current_app.logger.warning(error)
                raise Forbidden(description=description)

        request_csrf_token = request.headers.get(self._csrf_header_name, '')

        some_none = None in (request_csrf_token, server_csrf_token)
        if some_none or not safe_str_cmp(request_csrf_token, server_csrf_token):
            error = (REASON_BAD_TOKEN, request.path)
            error = u'Forbidden ({0}): {1}'.format(*error)
            current_app.logger.warning(error)
            raise Forbidden(description=REASON_BAD_TOKEN)
